
void topegCLASS::Loop()
{
/** @brief
 * Loop through a TOPEG DVCS root tree and convert to HepMC2 output
 */
   if (fChain == 0) return;
    static double leptonMass;
    static double ionMass;
    double px, py, pz, pp0, E0, EIon, Ep;
    double kx, ky, kz, k0, e0, ELepton;
    double BetaGamma;
    double tnthC = tan(CrossingTheta);
    //Setup HepMC2 output format
    printf("opening output file %s \n",outputfile);
    HepMC3::WriterAsciiHepMC2 hep_output(outputfile);
   Long64_t nentries = fChain->GetEntriesFast();
    printf("number of entries: %lld \n", nentries);
   Long64_t nbytes = 0, nb = 0;
   for (Long64_t jentry=0; jentry<nentries;jentry++) {
      Long64_t ientry = LoadTree(jentry);
       if (ientry < 0) break;
       nb = fChain->GetEntry(jentry);   nbytes += nb;
       /**
        * Load the full event, then globally Boost and Rotate to equivalent crossing kinematics
        */

       HepMC3::GenEvent * hep_evt = new GenEvent(HepMC3::Units::GEV, HepMC3::Units::MM);
       hep_evt->set_event_number(jentry);
       // Get Beam particles, in head-on kinematics
       if (jentry==0){
           leptonMass = dbPDG->GetParticle(PIDlBeam)->Mass();
           ionMass = dbPDG->GetParticle(PIDhBeam)->Mass();
           printf("entry %lld, electron %7.3f GeV,  Ion PID %d, Mass %8.4f GeV \n",
                  jentry, ElBeam, PIDhBeam, ionMass);

       }
       // Vertex (could be smeared)
       hep_4vec = HepMC3::FourVector(0.0,0.0,0.0,0.0); //(x,y,z,t)
       HepMC3::GenVertexPtr hep_vtx = make_shared<GenVertex>(hep_4vec);
       hep_evt->add_vertex(hep_vtx);
       //  Beam Lepton
       k0 = sqrt(ElBeam*ElBeam-leptonMass*leptonMass);
       hep_4vec = HepMC3::FourVector(0.0,0.0, -k0, ElBeam);
       HepMC3::GenParticle * hep_part = new HepMC3::GenParticle(hep_4vec, PIDlBeam);
       GenParticlePtr p1 = make_shared<GenParticle>(hep_4vec, PIDlBeam, beamStatus);
       hep_vtx->add_particle_in(p1);
       pp0 = sqrt(EhBeam*EhBeam-ionMass*ionMass);
       // Find equivalent ion momentum at crossing angle
       /* incorrect, deprecated
       pp = 2.0*pp0/cos(CrossingTheta) / ( 1.0+sqrt(1.0 + 2.0*pp0/(EhBeam+pp0) *tan(CrossingTheta)*tan(CrossingTheta)) );
       px = pp*sin(CrossingTheta)*cos(CrossingPhi);
       py = pp*sin(CrossingTheta)*sin(CrossingPhi);
       pz = pp*cos(CrossingTheta);
       EIon = sqrt(pp*pp+ionMass*ionMass);
       hep_4vec = HepMC3::FourVector(px,py,pz,EIon);
        */
       hep_4vec = HepMC3::FourVector(0.0,0.0,pp0,EhBeam);
       GenParticlePtr p2 = make_shared<GenParticle>(hep_4vec, PIDhBeam, beamStatus);
       hep_vtx->add_particle_in(p2);
       for (int ipart=0; ipart<Nb_part; ipart++) {
           px = part_px[ipart];
           py = part_py[ipart];
           pz = part_pz[ipart];
           Ep = part_e[ipart];
           hep_4vec= HepMC3::FourVector(px,py,pz,Ep);
           GenParticlePtr pOut = make_shared<GenParticle>(hep_4vec, part_id[ipart], scatStatus);
           hep_vtx->add_particle_out(pOut);
       }
       
       /** HepMC3 Rotate and boost all 4-vectors of an event!!
        *bool GenEvent::rotate(const FourVector&  delta): rotation of coordinates (passive)
        * by angle |delta| in sense of right hand rule around axis delta
        *bool GenEvent::boost(const FourVector&  delta):  boost opposite delta
        */

       /**
        * Convert head-on collision to equivalent crossing angle collision at angle
        * \f$ \theta_C  \f$.
        * Note this does not necessarily conserve the phase space volume of the original particle generation \\
        * Transverse boost \\
        *  Initial four vectors: Lepton: \f$  [e_0, 0, 0, -k_0] \f$ ,
        *  Ion:  \f$ [E_0, 0, 0, P_0] \f$
        * \f{align}{
        *  \beta_\perp\gamma = \frac{ 2 k_0 P_0}{E_0 k_0 + e_0 P_0}\left/\left\{
        *  1 + \sqrt{1 + \frac{4k_0 P_0 E_0 e_0 \tan\theta_C}{(E_0 k_0 + e_0 P_0)^2}}
        *  \right\} \right.
        *  \f}
        */
//       Print::listing(*hep_evt,2);
       e0 = ElBeam;
       k0 = sqrt(ElBeam*ElBeam-leptonMass*leptonMass);
       E0 = EhBeam;
       // pp0= sqrt(E0*E0-ionMass*ionMass);
       BetaGamma = ( 2.0*k0*pp0*tnthC/(E0*k0+e0*pp0) /
                    (1.0 + sqrt(1.0 +4.0*k0*pp0*E0*e0*tnthC*tnthC /((E0*k0+e0*pp0)*(E0*k0+e0*pp0))  ) ) );
       double ThBoost = atan(BetaGamma*e0/k0);
       double betaBoost = BetaGamma/sqrt(1.0+BetaGamma*BetaGamma);
       // Boost 4-vector
       hep_4vec = HepMC3::FourVector(-betaBoost*cos(CrossingPhi), -betaBoost*sin(CrossingPhi),0.0,0.0);
       if(!hep_evt->boost(hep_4vec)){
           printf("event %lld boost failed",jentry);
           break;
       }
       hep_4vec = HepMC3::FourVector(sin(CrossingPhi)*ThBoost, -cos(CrossingPhi)*ThBoost,0.0,0.0);
       if(!hep_evt->rotate(hep_4vec)){
           printf("event %lld rotation failed",jentry);
           break;
       }
 //      Print::listing(*hep_evt,2);
       hep_output.write_event(*hep_evt);
       hep_evt->clear();
       delete hep_evt;
   }
    hep_output.close();
}
